from typing import Any, List

try:
    import orjson as json
except ImportError:
    import json
from json.decoder import JSONDecodeError

from seito.monad.opt import none, Some, Empty, opt
from seito.monad.try_ import attempt_to


class JsObject(dict):
    def __init__(self, *args, **kwargs):
        super(JsObject, self).__init__(*args, **kwargs)

    def stringify(self, **kwargs: Any):
        return json.dumps(self, **kwargs)

    def __getitem__(self, item: str) -> Some[Any] | Empty:
        try:
            v = super(JsObject, self).__getitem__(item)
            return opt(v)
        except KeyError:
            return none

    def __getattr__(self, item: str) -> Some[Any] | Empty:
        return self[item]

    def __setattr__(self, key, value):
        self[key] = value

    def __str__(self):
        return json.dumps(self)


def js(*args, **kwargs):
    return JsObject(*args, **kwargs)


def parse(string: str | bytes | bytearray, *, response_class=None, **kwargs: Any):
    value = json.loads(string, **kwargs)
    is_list = isinstance(value, List)
    if response_class is None:
        if is_list:
            return [js(val) for val in value]
        return js(value)
    if is_list:
        return [response_class(**val) for val in value]
    return response_class(**value)


try_parse = attempt_to(errors=(JSONDecodeError,))(
    parse
)



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, TypeVar, Generic, Callable, NoReturn

from pampy import _ as __
from pampy.pampy import match_dict as pampy_dict_matcher

from seito.monad.func import identity


class EmptyError(ValueError):
    ...


class MatchError(TypeError):
    ...


def apply(f: Callable[..., Any] | Any = identity, *args: Any, **kwargs: Any) -> Any:
    if callable(f):
        return f(*args, **kwargs)
    return f


_ = __


class When:
    def __init__(self, value):
        self.value = value
        self.action = None

    def then(self, f):
        self.action = f
        return self

    def __repr__(self):
        return repr(self.value)


when = When


@dataclass
class Default:
    def __init__(self):
        self.action = None

    def __rshift__(self, other) -> "Default":
        self.action = other
        return self


default = Default


class Option(ABC):
    @abstractmethod
    def get(self) -> Any:
        ...  # pragma: no cover

    @abstractmethod
    def is_empty(self) -> bool:
        ...  # pragma: no cover

    @abstractmethod
    def or_else(self, obj: Callable[..., Any] | Any, *args: Any, **kwargs: Any) -> Any:
        ...  # pragma: no cover

    @abstractmethod
    def or_if_falsy(
            self, obj: Callable[..., Any] | Any, *args: Any, **kwargs: Any
    ) -> Any:
        ...  # pragma: no cover

    @abstractmethod
    def or_none(self) -> Any:
        ...  # pragma: no cover

    @abstractmethod
    def or_raise(self, exc: Exception):
        ...  # pragma: no cover

    @abstractmethod
    def map(self, func: Callable[..., Any], *args: Any, **kwargs: Any) -> "Option":
        ...  # pragma: no cover

    @abstractmethod
    def __iter__(self):
        ...  # pragma: no cover

    @abstractmethod
    def __getattr__(self, name: str) -> Any:
        ...  # pragma: no cover

    @abstractmethod
    def __call__(self, *args: Any, **kwargs: Any):
        ...  # pragma: no cover

    @abstractmethod
    def __str__(self) -> str:
        ...  # pragma: no cover

    def __rshift__(self, other) -> When:
        return when(self).then(other)

    def match(self, *whens: When | Default):
        for w in whens:
            if isinstance(w, when):
                if w.value.__class__ == self.__class__:
                    match_dict, self_dict = w.value.__dict__, self.__dict__
                    is_a_match, extracted = pampy_dict_matcher(match_dict, self_dict)
                    if not is_a_match:
                        continue
                    if extracted:
                        return apply(w.action, *extracted)
                    return apply(w.action)
            else:
                return apply(w.action)
        raise MatchError(f"No default guard found, enable to match {self}")


T = TypeVar("T")
M = TypeVar("M")


@dataclass
class Some(Generic[T], Option):
    _under: T

    def get(self) -> T:
        return self._under

    def is_empty(self) -> bool:
        return False

    def or_else(self, obj: Callable[..., Any], *args: Any, **kwargs: Any) -> T:
        return self._under

    def or_if_falsy(
            self, obj: Callable[..., Any], *args: Any, **kwargs: Any
    ) -> T | Any:
        return self._under or apply(obj, *args, **kwargs)

    def or_none(self) -> T:
        return self._under

    def or_raise(self, exc: Exception) -> T:
        return self._under

    def map(
            self, f: Callable[[T, ...], M], *args: Any, **kwargs: Any
    ) -> "Some[M] | Empty":
        inst = self
        while isinstance(inst._under, Option):
            inst = inst._under
        return opt(apply(f, inst._under, *args, **kwargs))

    def __iter__(self):
        under = self._under
        while under is not None:
            if isinstance(under, Option):
                under = under._under
            else:
                yield under
                under = None

    def __getattr__(self, name: str) -> "Some[Any] | Empty | Callable[..., Any]":
        try:
            attr = getattr(self._under, name)
        except AttributeError:
            return none
        if callable(attr):
            def wrapper(*args: Any, **kwargs: Any) -> "Some | Empty":
                return opt(attr(*args, **kwargs))

            return wrapper
        return opt(attr)

    def __call__(self, *args: Any, **kwargs: Any) -> "Some[T]":
        return self

    def __str__(self) -> str:
        return f"<Some {str(self._under)}>"


@dataclass
class Empty(Option):
    def get(self) -> NoReturn:
        raise EmptyError("Option is empty")

    def is_empty(self) -> bool:
        return True

    def or_else(self, obj: Callable[..., M] | M, *args: Any, **kwargs: Any) -> M:
        return apply(obj, *args, **kwargs)

    def or_if_falsy(self, obj: Callable[..., M] | M, *args, **kwargs) -> M:
        return apply(obj, *args, **kwargs)

    def or_none(self) -> None:
        return None

    def or_raise(self, exc: Exception) -> NoReturn:
        raise exc

    def map(self, func, *args, **kwargs) -> "Empty":
        return self

    def __iter__(self):
        return self

    def __next__(self):
        raise StopIteration()

    def __getattr__(self, name: str) -> "Empty":
        return self

    def __call__(self, *args: Any, **kwargs: Any) -> "Empty":
        return self

    def __str__(self) -> str:
        return "<Empty >"


E = TypeVar("E", bound=Exception)


@dataclass
class Err(Empty, Generic[E]):
    _under: E

    def __str__(self):
        return f"<Err {repr(self._under)} >"

    def unwrap(self):
        return self._under

    recover_with = Empty.or_else


def option(value: Any) -> Some[Any] | Empty:
    if isinstance(value, Exception):
        return Err(value)
    return none if value is None else Some(value)


def opt_from_call(f, *args, **kwargs):
    exc = kwargs.pop("exc", Exception)
    try:
        return opt(f(*args, **kwargs))
    except exc:
        return Err(exc)


# aliases
none = nope = empty = Empty()
opt = option
err = Err



import asyncio
from asyncio import iscoroutinefunction
from typing import Awaitable, Callable, Any

from aflowey import aflow, partial, CANCEL_FLOW, F
from loguru import logger

from seito.monad.opt import T, Option, identity, opt, When, Default, Some, Empty, Err


async def _exec(function, *a: Any, **kw: Any) -> Any:
    current_result = function(*a, **kw)
    while asyncio.iscoroutine(current_result):
        current_result = await current_result
    if isinstance(current_result, Option):
        return current_result
    if callable(current_result):
        return await _exec(current_result)
    return current_result

def breaker_wrapper(f):
    async def w(*args, **kwargs):
        value = await _exec(f, *args, **kwargs)
        if value is None:
            return CANCEL_FLOW
        return value

    return w


class AsyncOption(Option):
    def __init__(self, aws: Awaitable[Any] | Any, *args: Any, **kwargs: Any) -> None:
        self._under = aws
        self.args = args
        self.kwargs = kwargs
        self._mappers = []

    async def _execute(self):
        init_f = breaker_wrapper(
            partial(self._under, *self.args, **self.kwargs)
        )
        # currently, a bug exists in aflowey, have to check the return value
        # of the first call
        try:
            value = await init_f()
        except Exception as e:
            return e

        if value is CANCEL_FLOW or value is None:
            return value

        # option can not be used in pipeline
        if isinstance(value, Option):
            return value

        flow = aflow.from_args(value) >> identity
        for (f, a, kw) in self._mappers:
            flow = flow >> breaker_wrapper(partial(f, *a, **kw))
        try:
            return await flow.run()
        except Exception as e:
            logger.error(e)
            return e

    async def _execute_or_clause(self, or_f, *args, **kwargs):
        if iscoroutinefunction(or_f):
            return await or_f(*args, **kwargs)
        elif callable(or_f):
            return or_f(*args, **kwargs)
        return or_f

    async def _execute_or_clause_if(self, predicate, or_f, *args, **kwargs):
        value = await self._execute()
        if predicate(value):
            return await self._execute_or_clause(or_f, *args, **kwargs)
        return value

    def map(self, f, *args, **kwargs):
        self._mappers.append((f, args, kwargs))
        return self

    async def get(self) -> Any:
        result = await self._execute()
        return opt(result).get()

    async def or_none(self) -> Any:
        value = await self._execute()
        return opt(value).or_none()

    async def or_else(
        self, or_f: Callable[..., Any] | T, *args: Any, **kwargs: Any
    ) -> T:
        predicate = (
            lambda value: value is None
            or value is CANCEL_FLOW
            or isinstance(value, Exception)
        )
        return await self._execute_or_clause_if(predicate, or_f, *args, **kwargs)

    async def or_if_falsy(
        self, or_f: Callable[..., Any] | Any, *args: Any, **kwargs: Any
    ) -> Any:
        return await self._execute_or_clause_if(
            lambda value: not value, or_f, *args, **kwargs
        )

    async def or_raise(self, exc: Exception):
        value = await self._execute()
        return opt(value).or_raise(exc)

    async def is_empty(self) -> bool:
        value = await self._execute()
        return opt(value).is_empty()

    def __aiter__(self):
        class Aiter:
            def __init__(self, exc):
                self.i = 0
                self.exc = exc

            async def __anext__(self):
                if self.i == 1:
                    raise StopAsyncIteration()
                value = await self.exc()
                as_opt = opt(value)
                for val in as_opt:
                    self.i += 1
                    return val
                else:
                    raise StopAsyncIteration()

        return Aiter(self._execute)

    async def __call__(self, *args: Any, **kwargs: Any):
        value = await self._execute()
        return opt(value)(*args, **kwargs)

    def __str__(self) -> str:
        return f"<AsyncOption {self._under}>"

    def __iter__(self):
        raise NotImplementedError() # pragma: no cover

    def __getattr__(self, name: str) -> Any:
        raise NotImplementedError() # pragma: no cover

    async def match(self, *whens: When | Default):
        result = await self._execute()
        return opt(result).match(*whens)


aopt = AsyncOption



from enum import Enum
from typing import Type, Any

from aiohttp import ClientSession
from loguru import logger

from seito.json import try_parse
from seito.monad.opt import Err, Some


class HttpMethods(str, Enum):
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"
    PATCH = "PATCH"
    OPTIONS = "OPTIONS"


class ContentType:
    value = "Content-Type"
    APPLICATION_JSON = "application/json"


class NetworkError(Exception):
    ...


class HttpException(NetworkError):
    def __init__(self, code: int, detail: str):
        self.code = code
        self.detail = detail


class NetworkException(NetworkError):
    ...


async def request(
    method: HttpMethods,
    url,
    /,
    *,
    session: ClientSession,
    response_class: Type[Any] | None=None,
    **kwargs: Any,
) -> Some | Err[NetworkError]:
    try:
        async with session.request(method.value, url, **kwargs) as resp:
            content_type = resp.headers.get(ContentType.value)
            is_json = content_type == ContentType.APPLICATION_JSON
            resp_as_text = await resp.text()
            if resp.status >= 400:
                return Err(HttpException(400, resp_as_text))
            if is_json:
                if response_class is not None:
                    return try_parse(resp_as_text, response_class=response_class)
                else:
                    return try_parse(resp_as_text)
            return Some(resp_as_text)
    except Exception as e:
        logger.error(e)
        return Err(NetworkError())
